#include "main.h"
#include "monitor.h"
#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <string.h>
#include <ctype.h>

#define CRLF "\r\n"		// carriage return + line feed
#define MAX_CMDLEN 14
#define SPACE ' '
#define MAXMEM 16384
#define BYTES_LINE 8
#define PRINT_DELAY 30	// ms delay per uart_putc

static uint16_t current = 0x800;	// default memory address
static uint8_t base = 16;		// number base for memory dump
static uint8_t SDst;	// SD Card Status

// private prototypes
void ser9600_8n1(void);
uint8_t uart_putc(uint8_t c, uint8_t delay);
void uart_puts (char *s);
uint8_t uart_getc(void);
void print(char *s);
void process_command(uint8_t *buf);
uint8_t atohex(uint8_t *s, uint8_t L, uint16_t *x);	// ASCII to hex
void helptext(void);
void dispmem(uint8_t a1L, uint8_t a2L, uint16_t a1, uint16_t a2, uint8_t base);
void ntoa(uint16_t n, uint8_t *buf, uint8_t base, uint8_t digits);
void LoadProg(uint16_t, uint8_t);
void SaveProg(uint16_t, uint16_t);

void monitor(void)
{
uint8_t pos = 0;
uint8_t err;
uint8_t c;	// char read from USART
uint8_t cmd[20];	// command buffer for monitor
	ser9600_8n1();	// show RS232 parameter in K1000 display
	print("");
	print("**** K1000 Monitor V0.3 ****");
	print("     (C) 2012 M.Berger");
	print("     press ? for command overview");
	print("");
	if ((SDst = mmc_init()) != 0) print("SD-CARD not initialized"); // assignment inside test
	uart_putc('>',0);	// command prompt
	do	{	// monitor main loop
	    err = 0;	// clear error flag
		pos = 0;	// reset position pointer
		if(ix != ox) {		// if a key was pressed (yet unspecified which)
			do {
				if (ix != ox) {
					c = toupper(buf[ox++]);	// get char and increment pointer
					if ((c >=' ') && (pos<MAX_CMDLEN)){
						cmd[pos] = c;	// store char
						uart_putc(c, 0);		// echo to terminal
						pos++;			
					}			
				}						
			} while (c != '\r'); // wait for <RETURN> to complete input
			cmd[pos] = '\0';	// terminate string
			uart_puts(CRLF);	// issue CR LF
			// now process command
			process_command(cmd);
			if (sp_key) uart_putc('>', 0);	// next command prompt
		}				
	} while (sp_key); // end of main loop

	// UCSR0B = 0b00000000;	// disable TX and RX
}

void ser9600_8n1(void)
{
uint8_t i;
	for (i=0;i<16;i++) dispBuffer[i] = 0;
	dispBuffer[4] = 182;	// S
	dispBuffer[3] = 158;	// E
	dispBuffer[2] = 10;		// r
	
	dispBuffer[0] = 246;	// 9
	dispBuffer[15] = 190;	// 6
	dispBuffer[14] = 252;	// 0
	dispBuffer[13] = 252;	// 0
	
	dispBuffer[12] = 254;	// 8
	dispBuffer[11] = 42;	// n
	dispBuffer[10] = 96;	// 1
		
}

void uart_init(void)
{
	_delay_ms(100);
	UBRR0H = 0;
	UBRR0L = 119; // 9600 Baud for F_CPU = 18.432 MHz
	UCSR0B = (1<<RXEN0) | (1<<TXEN0) | (1<<RXCIE0);	// enable interrupt for RxD
	UCSR0C = (3<<UCSZ00);	// 8N1 format
	
	uart_putc(RESET, PRINT_DELAY);		// send a RESET to the printer
	_delay_ms(100);
}

// send one character
uint8_t uart_putc(uint8_t c, uint8_t delay)
{

    while (!(UCSR0A & (1<<UDRE0)))  /* warten bis Senden moeglich */
	{
	}	                       
    UDR0 = c;                      /* sende Zeichen */
	if (delay) _delay_ms(delay);	// give printer a little bit extra time ;)
    return 0;
}
 
// send a string
void uart_puts (char *s)
{
    while (*s)
    {   /* so lange *s != '\0' also ungleich dem "String-Endezeichen(Terminator)" */
        uart_putc(*s, 0);
        s++;
    }
}

// receive one character
uint8_t uart_getc(void)
{
    while (ix == ox) {
		// wait for incoming char
		}
    return toupper(buf[ox++]);   // Zeichen an Aufrufer zurueckgeben
}

void print(char *s)
{
	uart_puts(s);
	uart_puts(CRLF);
}

void process_command(uint8_t *buf)
{
uint8_t cmd;
uint8_t err;
uint8_t *here;
uint8_t *p;
uint8_t arg1[10], arg2[10];
uint16_t a1,a2;	// numeric values
uint8_t a1L,a2L;	// length and existing flag
    err = 0;	// clear error flag
    // evaluate input
    here = buf;
    a1L = a2L = 0;  // assume no args exist 

    while (*here == SPACE) here++;  // skip leading spaces
    cmd = toupper(*(here++));	// get cmd and ++ pointer
    if (NULL == (p = strchr("DHLMOPRSWX?",cmd))) err = -1;
 
    while (*here == SPACE) here++;  // skip spaces if any
    if (*here > SPACE) { // arg1 exists
      while (*here > SPACE) {
        arg1[a1L] = toupper(*here);
        a1L++; here++;
        arg1[a1L] = '\0';	// string terminator
      }
    }
    if (a1L) err |= atohex(arg1,a1L,&a1);	

    while (*here == SPACE) here++;  // skip spaces if any
    if (*here > SPACE) { // arg2 exists
      while (*here > SPACE) {
        arg2[a2L] = toupper(*here);
        a2L++; here++;
        arg2[a2L] = '\0';	// string terminator
      }
    }
    if (a2L) err |= atohex(arg2,a2L,&a2);
    if (err) print("press ? for help");
    else {	// execute command
     switch(cmd) {
     case 'D':	// Hex mode
      base = 10;
      print("Decimal Mode");
      break;
     case 'H':	// Hex mode
      base = 16;
      print("Hex Mode");
      break;
	 case 'L':
	  LoadProg(a1,a1L);
	  break;
     case 'M':	// display Memory
      dispmem(a1L,a2L,a1,a2,base);
      break;
     case 'O':	// Octal Mode
      base = 8;
      print("Octal Mode");
      break;
	 case 'P':	// put Byte in memory
	  KRam[a1] = (uint8_t) a2 & 0xFF;
	  break;
	 case 'S':
	  SaveProg(a1,a2);
	  break;
     case 'X':	// eXit
	  sp_key = 0;		// flag for monitor exit
      break;  
     case '?':	// help
      helptext();
      break;
     }  	
    } 
	
}

/* -------------------------------
  atohex, converts ascii to hexadecimal
*/
uint8_t atohex(uint8_t *s, uint8_t L, uint16_t *x)
{
uint8_t err = 0;
uint8_t i, dig;
  if (L>4) err = -1;  // allow no more than 4 digits	
  else {
    *x = 0;
    for (i=0;i<L;i++) {
      dig = *(s++) - 48;	// ASCII to numeric
      if (dig > 9) dig -= 7;
      if (dig > 15) err = -1;	
      *x += dig << (L-i-1)*4;
    }
  if (*x >= MAXMEM) err = -1; 
  }
  return err;
} // end of atohex

void helptext(void)
{
  print("Command Overview");
  print("");
  print(" o                  ... set number base 8");
  print(" d                  ... set number base 10");
  print(" h                  ... set number base 16");
  print(" l [addr]           ... load program via serial line");
  print(" s addr1 addr2      ... save program via serial line");
  print(" m [addr1 [addr2]]  ... display memory dump");
  print(" p addr byte        ... put a byte into memory");
  print(" x                  ... exit from monitor");
} // end of helptext

void dispmem(uint8_t a1L, uint8_t a2L, uint16_t a1, uint16_t a2, uint8_t base)
{
uint16_t aa,ab;
uint8_t buf[6];	// buffer for number conversion
uint16_t i,j;
	if (a1L) aa = a1;
        else aa = current;
	if (a2L) ab = a2;
	else ab = aa+10 * BYTES_LINE;	// default 10 lines output
	for (i=aa;i<ab;i+=BYTES_LINE) {
	  ntoa(i,buf,16,4);
	  uart_puts(buf); uart_puts(": ");	// address + ": "
      for (j=0;j<BYTES_LINE;j++) {
        if (base<16) ntoa(memory(i+j),buf,base,3);	// DEC and OCT: 3 digits
        else ntoa(memory(i+j),buf,base,2);
        uart_puts(buf); uart_putc(' ', 0);	// Byte followed by space
      }
          print("");	// CR LF
	}
        current = i;
} // end of dispmem

void ntoa(uint16_t n, uint8_t *buf, uint8_t base, uint8_t digits)
{
uint8_t c[] = "0123456789ABCDEF";
uint8_t *ptr;
uint8_t i;
  ptr = buf + digits;
  *(ptr--) = '\0';
  for (i=0;i<digits;i++) {
    *(ptr--) = c[n % base];
    n = n / base;
  }
} // end of ntoa

void LoadProg(uint16_t addr, uint8_t a1L)
{
uint8_t nib1,nib2;
uint8_t c,i;
uint8_t hi = 1;		// flag for high nibble
uint16_t addr2 = 0;	// start addr from file as proposal
	PORTA = 0;	// clear current display digit
    do {
		c = uart_getc();
	} while (c != '%');		// wait for address prefix
	for (i=0;i<4;i++) {
		c = uart_getc();
		c -=48;
		if (c > 9) c-= 7;	// convert to number
		addr2 = (addr2 << 4) + c; 
	}	
	if (!a1L) addr = addr2;	// explicit address overrides addr from file
	do {
		c = uart_getc();
		if (c=='.') break;			// '.' is END OF FILE signal
		c -=48;
		if (c > 9) c-=7;	// convert to number
		if (c<16) {			// only process valid numbers
			if (hi) KRam[addr] = c << 4;
			else {
				KRam[addr++] |= c;
				uart_putc('.',0);	// show that we are still alive!
			}			
			hi = 1-hi;		// invert hi-lo nibble flag
		}
	} while (1);
	print("");
	print("READY.");	
}

void SaveProg(uint16_t addr1, uint16_t addr2)
{
uint16_t byte;
uint8_t buf[6];
uint8_t cnt=0;
	uart_putc('%',0);
	ntoa(addr1,buf,16,4);
	print(buf);		// start addr. + CRLF
	while (addr1 < addr2) {
		byte = memory(addr1++);	// increment addr1 afterwards
		ntoa(byte,buf,16,2);
		uart_puts(buf);
		cnt++;
		if (cnt > 15) {cnt=0;print(""); }	// 16 Byte per Line
	}
	print(".");
}

// ----------- begin of K1003 printer code ---------------

void PrintLine(void)
{
uint16_t i;
uint8_t c;
	for (i=0x085F;i>=0x0850;i--) {
		c = KRam[i];	// get original K1003 char
		KRam[i] = 95;	// reset to: K1000 code for space 
		c = pgm_read_byte(&K1000ZS[c]);	// convert to ASCII
		uart_putc(c, PRINT_DELAY);		
	}		
	uart_puts(CRLF);	// New Line

	// *********
	while (StopTransmission)		// XOFF/XON Protokoll
	{
	}
	// *********
	
}
